import logging

import zmq
from zmq.eventloop.zmqstream import ZMQStream

from .scope import Scope
from .serialization import (
    serialize_tfw_msg,
    deserialize_tfw_msg,
    with_deserialize_tfw_msg
)

LOG = logging.getLogger(__name__)


class ZMQDownlinkConnector:
    def __init__(self, connect_addr):
        self._zmq_sub_socket = zmq.Context.instance().socket(zmq.SUB)
        self._zmq_sub_socket.setsockopt(zmq.RCVHWM, 0)
        self._zmq_sub_socket.connect(connect_addr)
        self._zmq_sub_stream = ZMQStream(self._zmq_sub_socket)

    def subscribe(self, *keys):
        for key in keys:
            self._zmq_sub_socket.setsockopt_string(zmq.SUBSCRIBE, key)

    def unsubscribe(self, *keys):
        for key in keys:
            self._zmq_sub_socket.setsockopt_string(zmq.UNSUBSCRIBE, key)

    def register_callback(self, callback):
        callback = with_deserialize_tfw_msg(callback) if callback else None
        self._zmq_sub_stream.on_recv(callback)

    def recv_message(self, *, block=True):
        if self._zmq_sub_stream.receiving():
            raise RuntimeError('Synchronous recv() called while a callback is registered!')
        flags = 0 if block else zmq.NOBLOCK
        try:
            return deserialize_tfw_msg(*self._zmq_sub_socket.recv_multipart(flags))
        except zmq.ZMQError:
            raise IOError("No data available to recv!")

    def close(self):
        self._zmq_sub_stream.close()


class ZMQUplinkConnector:
    def __init__(self, connect_addr):
        self._zmq_push_socket = zmq.Context.instance().socket(zmq.PUSH)
        self._zmq_push_socket.setsockopt(zmq.SNDHWM, 0)
        self._zmq_push_socket.connect(connect_addr)

    def send_message(self, message, scope=None, intent=None):
        if 'scope' not in message:
            message['scope'] = Scope.ZMQ.value
        if scope is not None:
            message['scope'] = scope.value
        if intent is not None:
            message['intent'] = intent.value
        self._zmq_push_socket.send_multipart(serialize_tfw_msg(message))

    def close(self):
        self._zmq_push_socket.close()


class ZMQConnector(ZMQDownlinkConnector, ZMQUplinkConnector):
    def __init__(self, downlink_connect_addr, uplink_connect_addr):
        ZMQDownlinkConnector.__init__(self, downlink_connect_addr)
        ZMQUplinkConnector.__init__(self, uplink_connect_addr)

    def close(self):
        ZMQDownlinkConnector.close(self)
        ZMQUplinkConnector.close(self)
